!*****************************************************************************

!  MPFUN20-Fort: A thread-safe arbitrary precision computation package
!  Global data definition module (module MPFUNA)

!  Revision date:  9 Dec 2020

!  AUTHOR:
!     David H. Bailey
!     Lawrence Berkeley National Lab (retired) and University of California, Davis
!     Email: dhbailey@lbl.gov

!  COPYRIGHT AND DISCLAIMER:
!    All software in this package (c) 2020 David H. Bailey.
!    By downloading or using this software you agree to the copyright, disclaimer
!    and license agreement in the accompanying file DISCLAIMER.txt.

!  PURPOSE OF PACKAGE:
!    This package permits one to perform floating-point computations (real and
!    complex) to arbitrarily high numeric precision, by making only relatively
!    minor changes to existing Fortran-90 programs.  All basic arithmetic
!    operations and transcendental functions are supported, together with several
!    special functions.

!    In addition to fast execution times, one key feature of this package is a
!    100% THREAD-SAFE design, which means that user-level applications can be
!    easily converted for parallel execution, say using a threaded parallel
!    environment such as OpenMP.  There are NO global shared variables (except
!    static compile-time data), and NO initialization is necessary unless
!    extremely high precision (> 20,000 digits) is required.

!  DOCUMENTATION:
!    A detailed description of this package, and instructions for compiling
!    and testing this program on various specific systems are included in the
!    README file accompanying this package, and, in more detail, in the
!    following technical paper:
   
!    David H. Bailey, "MPFUN2020: A new thread-safe arbitrary precision package," 
!    available at http://www.davidhbailey.com/dhbpapers/mpfun2020.pdf.

!  DESCRIPTION OF THIS MODULE (MPFUNA):
!    This module contains some global parameter definitions, including data
!    statements for log(2) and pi.  See documentation for details.

module mpfuna

!----------------------------------------------------------------------------

!   Integer(4) constants:

!   Name     Default   Description
!   mpiknd        8    Kind parameter for long integers.
!   mpiknd2      16    Kind parameter for quad integers.
!   mpndpw       14    Largest n such that 10^n <= 2^53, the max of DP integers.
!   mpldb         6    Logical device number for output of error messages.
!   mpnbt        60    Number of mantissa bits per long integer in MP numbers.
!   mpnstr     2048    Maximum length of certain input character strings.
!                        See usage in mpinp of module MPFUNC.
!   mpoutl       80    Length of output lines.  See usage in mpout of MPFUNC.
!   mprknd        8    Kind parameter for double precision floating-point.
!   mprknd2      16    Kind parameter for quad precision floating-point.

!   Integer(8) constants:
!   Name     Default   Description
!   mpbdx      2^60    2^mpnbt, the radix for MP numbers.

!   Real(8) constants:
!   Name     Default   Description
!   mpb13x     2^13    Constant for checking the 40-bit restriction.
!                        See usage in mpdmc40, mpmuld40 and mpdivd40 of MPFUNB.
!   mpdpw  Log10(2^48) DP approximation to number of digits per mantissa word.
!   mpexpmx   2^31/48  Largest permissible exponent, corresponding to a maximum
!                        binary exponent of 2^31.

implicit none
integer, public:: mpiknd, mpiknd2, mpndpw, mpldb, mpnbt, mpnstr, mpoutl, &
  mprknd, mprknd2
parameter (mpiknd = 8, mpiknd2 = 16, mpndpw = 14, mpldb = 6, mpnbt = 60, &
  mpnstr = 2048, mpoutl = 80, mprknd = 8, mprknd2 = 16)
integer (mpiknd), public:: mpbdx
parameter (mpbdx = 2_8**mpnbt)
real (mprknd), public:: mpdpw, mpexpmx, mprdfz, mpb13x
parameter (mpdpw = 18.061799739838871713d0, &
  mpexpmx = 2.d0**31 / mpnbt, mprdfz = 0.5d0**50, mpb13x = 2.d0**13)

!  These are for debug and timing (for internal use only):

integer, dimension(10):: mpdeb = 0
real (mprknd), dimension(10):: mptm = 0.d0

!  This data is for the extra-high precision routines.

!  The following data arrays contain the values of log(2) and pi and 1/sqrt(2) to
!  over 20,000-digit precision (by default).  If more than this level are needed,
!  subroutine mpinitran (in module MPFUND) must be called to initialize data
!  for transcendental functions.  Alternatively, if you are using the
!  high-level interface in module MPFUNF, calling mpinit will do this.
!  The arrays below have sufficient space to support precision levels up to
!  1.8 million decimal digits.

!  ALL lines from this point on, except for the "end module" statement at the end,
!  are generated by the compcon.f90 program, available from the author.

integer, public:: mplconx
integer, private:: i 
parameter (mplconx =    100000)
integer (mpiknd), public:: mplog2con(0:mplconx-1), mppicon(0:mplconx-1)

data (mplog2con(i), i =     0,   199) / &
                1117_8,                1111_8,                1111_8,                  -1_8, &
  799144290325165978_8,  849460365225095926_8,  789272551555017099_8,  445163727010630570_8, &
 1126672294557648606_8,  841214723756325706_8, 1126376059045602140_8,   87856769335245609_8, &
   78421027581422753_8,  441998077035465690_8,   77431747161857705_8,  847716001976624495_8, &
  724735828797208364_8, 1047000727954180853_8, 1007236022851019910_8,  387205694960493474_8, &
  982586796677528924_8,  434585037786799099_8,  650225694886395477_8,  117918596040883230_8, &
 1126277945069789504_8,   50549047534515675_8,  619542214527054635_8,  322514470696668773_8, &
  664620783330501088_8,  851697140255516315_8,  230622882617010631_8, 1119810899738232653_8, &
  819679114870534340_8,  579367960721566869_8,  420268095221250522_8, 1022665503503576241_8, &
   34048807728765745_8,  819462395931944129_8,  627356470626432600_8,  377587242175372002_8, &
  522106768838125817_8,  475187583109223216_8,  790533142032307527_8,  224296586613789649_8, &
  155555536112645399_8,  575281007820180060_8,  142236722257074943_8,  956006791660920466_8, &
  156368997891986466_8,  109192266016426251_8,  842577403738103587_8,  495443378921047737_8, &
   25916096063257903_8,  676124210408125699_8,    9013857688257425_8,  724049519265889978_8, &
  563040853642300494_8,  657662912324494039_8,  248706839240402313_8,    8554658232026208_8, &
  992172116253532354_8,  658673254866731369_8,  494591850600231810_8,  503735489371114161_8, &
   79863948294375794_8,  923673311921914476_8,  320328893528628229_8, 1112244984639079854_8, &
  931320635882439481_8,  977114157202401929_8,  139338565954472227_8,  526383901974710490_8, &
  782931846310331300_8,  701732238205853066_8,   63683354286319583_8,  673434243579618325_8, &
  175548377669993545_8, 1146437748375563034_8,  448790188173634227_8,  222398369492404602_8, &
  260337464341295972_8, 1106630893137309678_8,  495542257580167279_8,  796140836195131404_8, &
  384125646372631712_8,    3563472316330036_8,  129446520001227935_8,  669307132983046785_8, &
  360821498254194255_8,  123711642274902542_8,  216024568108898111_8,  778102238170547600_8, &
  222184626363311026_8,  144718806408174580_8, 1100108707831828236_8,  979373091756646605_8, &
  818752698118121613_8,  398099411270273647_8, 1031791399515909930_8,  975666882559521296_8, &
  826302772131868796_8,  901957094641068366_8,  985468267140460917_8,  140636129575982693_8, &
  582443381339988004_8,  436965351657920632_8,  329690268516097450_8,  361755822954707778_8, &
  836048778640073208_8,  427683480088008861_8,  190472261838896532_8,   24640005490158421_8, &
  182284053086034962_8,  676578978331011955_8, 1034521786910346870_8,  166374410579214955_8, &
  468517924915632539_8,  463238016996657368_8,  914123998354776640_8,   57079355793199360_8, &
  247591949805732633_8,  369869227079445107_8,  828287238903563655_8,  139465334086976052_8, &
  231199710005878699_8,  293935834227235306_8, 1010239010349777194_8,  588109827225105535_8, &
  590056800218899881_8,   88342550679400581_8,  197664220934771921_8,  973390870419564439_8, &
 1022354071638101640_8,  327223931981822834_8,  673719157925286936_8,  499260709773167827_8, &
  881420100402606352_8,  749712803814218956_8,  267805591355837971_8,  342981175418154092_8, &
   87374448893336260_8,   35425629321345233_8, 1067200725895442892_8,  651708134106010702_8, &
  468020985336532970_8,  548746280850768044_8,  424306061864281399_8,  562675447305845169_8, &
  799083858332991258_8, 1034251353089834270_8,  506993838410332019_8,  479369101179680390_8, &
  621590972165683106_8,  863492129211160691_8, 1084316836670123534_8,  688503277056753841_8, &
  789791707159317783_8,  790387065527662561_8,  837134085022055354_8,   11277928609406305_8, &
  518861726845039286_8,  440464703345610292_8,  121718278379253877_8,  759832941642890638_8, &
 1076920608631819102_8,  717772455104956408_8,  568984710648663741_8,  205477930870397675_8, &
  269784892369263667_8,  849183475259682338_8,  616866334384698888_8,  692800234392969398_8, &
 1152230549462934572_8,  500503152127649118_8,  126245498771068966_8,  890929255574027784_8, &
 1060168635810577430_8,  595077510428262677_8,  731890091658661833_8,  986954251979502786_8, &
  527834251376216053_8,  416866295771433412_8,  413817092072915606_8,  679518426634438959_8, &
  367071260004420686_8, 1112439051266371949_8, 1067902753019227613_8,   88389281379550990_8, &
  512312274980148782_8,   77386272335476273_8,  150726669523121823_8,  614515173270362700_8, &
  692597806892870716_8, 1023928212596610006_8,  481491362894488537_8, 1044650816900609030_8, &
  625698959695448248_8,  575620053285159746_8,  983938581084548586_8,  427350812476028405_8/
data (mplog2con(i), i =   200,   399) / &
   26214601062588532_8,  155403891005965895_8,  257567160581744906_8,  248812513404632388_8, &
  322811653660106279_8,   98301930540390155_8,   95365878972252892_8,  318320877913390204_8, &
  231009522683768740_8,  362462804834654824_8,   51650025446364022_8,  941349327986594338_8, &
  858214689172523519_8,   97009468687083271_8,  533457405689974006_8,  961566524321172902_8, &
  748893129930498356_8,  739359652063325213_8,  376186840024265488_8,   81685434742660315_8, &
   18552275975468577_8,  759513350041907521_8, 1107918750620629381_8,  482623678831139631_8, &
  313587492061892516_8,  970119875325489495_8, 1152029418603015670_8,  292752718313153111_8, &
  729652919649859847_8,   59121243270169511_8,  883012910660207885_8,  911883276820725257_8, &
 1108060247203002448_8,  777499776365958319_8,  506579131957850575_8,  590390476447894226_8, &
  919038613562131972_8,  763797749262680806_8,  607116556032205078_8,  618832429999095046_8, &
  854095627383029666_8,  147420236702569455_8,  783122135413127828_8,  287429999575561501_8, &
  434511294326224078_8,  663224757235775484_8,  511524473030045632_8,  642446440584494050_8, &
  837936411221611485_8,  753614697706028886_8,  388231504997748244_8, 1145159457782254941_8, &
  569945831225802635_8,  886766057663486355_8,  975475112102166462_8, 1151236252682073041_8, &
 1148097606167348226_8,  394957217600550809_8,  771462837894568267_8,  998122784548308168_8, &
  503381597953737174_8,  322833980530501374_8,   87865924882473497_8,  714122648467357109_8, &
  752673846782047851_8,  250362701470943663_8,  205040300332387289_8,  977514481981426545_8, &
 1135119110138554444_8,  230249746973775135_8,  443104546631608384_8,  245356068942335423_8, &
 1030457727680559326_8,  192125873291591777_8,  283210844891684472_8,  555282640076409001_8, &
  998049996142846735_8,  591085771260265611_8, 1000408297633808930_8,  108251260765057132_8, &
  912112999405025082_8,  371580983923799882_8,  461139158578143836_8,  382447360157113668_8, &
  128210220471918635_8,  527076485125328577_8,  805699606227145128_8, 1116279669424834780_8, &
 1048040573894917156_8,  143576057199832021_8,  519513581632878400_8,  663887436479339473_8, &
  238951085900059841_8,  982094536536151870_8,  642992019531266513_8,  500695602980965738_8, &
  923792999635687764_8,  695738776817694694_8,  559009904195048011_8,  744630899780482644_8, &
  716978105369268336_8,  334185617048924857_8,  897074423348081097_8,   88807033510183842_8, &
  443586168665421195_8,  173781834081243802_8, 1113212896894046626_8,   27129351640591175_8, &
 1004081622300772378_8,  302418987647375213_8,  262044137353239909_8,  302105098921156801_8, &
 1121407493370256038_8,  673411397790718087_8,  832335897494446633_8,  800897952968847766_8, &
 1034238603232108545_8, 1096251132131054419_8, 1119274019573266152_8,   66405056886100240_8, &
  292019770913004824_8, 1059126497750198958_8,  976629425460808316_8,  581380674599560706_8, &
  620550942476223870_8,   88133070316701006_8,  411392517627898966_8,  441577778036813677_8, &
  867189274265345459_8,  759147152494146751_8,  575054746618095041_8,  901679738337181063_8, &
  808253674359980128_8,  792111553053836419_8,  857333890882998476_8,  316899465187571751_8, &
  677616162445074362_8, 1060780114976855105_8,  948117946844501581_8,  744702827442721304_8, &
  571866882866149613_8,  803644270804915261_8,  305782621818284514_8,  838164803311495193_8, &
   76151488926609854_8,  585522141654743831_8,   23008054201576536_8,  683442384478426519_8, &
 1074787843321260223_8, 1029376022591830482_8,  766420856548136942_8,  131778701675867495_8, &
  278548947942778515_8,  666628162075145654_8,  874713406086314666_8,  349223771627071413_8, &
  863589001508057903_8,  836916824891101261_8,  507350599049757394_8,  529082725265274023_8, &
  488608319470447175_8,  639663530610273294_8, 1025519810974580434_8,  607294123120628110_8, &
  428460088289282263_8, 1002420650540683836_8, 1106180566704768674_8,  558457124115438900_8, &
  366389169228121681_8, 1147442990070156565_8, 1082949516233001151_8,  986794671236436018_8, &
  278667451809049587_8, 1145981525557924905_8,  981544739076106474_8,  873732650238978614_8, &
  832928290735410628_8,  231619633199686161_8,   61933911224412632_8,  402116636713115237_8, &
  423729562159488429_8,  501843926785311192_8,  287033390816431576_8,  593396548481295684_8, &
  772218935421849099_8,  276532313064683585_8,  735268998693803397_8,  700729010767290589_8, &
 1085540102245623045_8,  958624864118475035_8,  249245384339645008_8, 1070855605738532307_8, &
  901005772729046013_8, 1024208502531112131_8,  807539004513036781_8,  809172551059996491_8, &
  810315367126421240_8, 1138624523581978994_8,  877927850664600021_8,  144027399777300786_8/
data (mplog2con(i), i =   400,   599) / &
  436526319822532119_8,  668467541211163834_8,  591122951432388562_8,  394235439063829825_8, &
  540233151069418882_8,  209615697160396555_8, 1005681268146132044_8,  687222575101143666_8, &
  451412544336235979_8,  658864515524322831_8,  376552379867535592_8,  415581362118889335_8, &
  631151753035475428_8,  892195505652087600_8,  450266715286191998_8,   92477002138264229_8, &
  421316017955593964_8,  257222533777835219_8, 1055621578801905721_8,  912369946434415798_8, &
  169343746217759468_8,  628221467855807740_8, 1088954373021956749_8,  888583260959617556_8, &
  232422717944375095_8,  570456111554037021_8,  102975415716891560_8,  768522359655049862_8, &
  905038678507803004_8,  408026425344761149_8,  742177891914447247_8,  336616977292951984_8, &
  959202627277445654_8,  677760873048061818_8,  975052682095487499_8,  679627407719411960_8, &
  586433832718688904_8,  156616636098739056_8,   16565359989183682_8, 1137052670363971832_8, &
  596773424405546043_8,  517530388650724475_8,  709570640198686087_8,  502148477400685260_8, &
  431205268083860841_8,  466390995821334447_8, 1004921374603432691_8,  939637548149927839_8, &
 1119451909969934809_8,  184118174201989108_8,  487998085446249452_8,  854513241285781248_8, &
 1022111897304564526_8,  660148626420111644_8, 1000257180853925634_8, 1050362794200386847_8, &
    2372508557559026_8,  756765172108950920_8,  719995798411006350_8,  440921952231799691_8, &
 1031634920328574259_8,  764158649254010567_8,  436510647072944019_8,  487342854280380908_8, &
   25353863464416841_8,  495411585321713556_8,   53726128399531960_8,  672623234073483761_8, &
   25409158502286721_8, 1101328548529495890_8,  468786814663215870_8,  984409316581227395_8, &
  562109035928892940_8,  132923348378120938_8,   10529244561450124_8,  584075375476198028_8, &
  435441147016303507_8, 1121389413910576846_8, 1119116310999176187_8,  580812415191731375_8, &
  607145523617859515_8,  138040645243900528_8,  993952911971738242_8,  586315428164924197_8, &
  693995277950933242_8,  450345089478105356_8,  778527095503729870_8,  765382812727126450_8, &
  272815728916914978_8,  110271630373740955_8,  104310950927544874_8, 1025406195559591999_8, &
  908377393288874436_8,  949114445089141098_8,  744104746054708323_8,  849571492676452013_8, &
  792063177787883245_8,   97980040087077128_8,  366600859212710196_8, 1046827633420619518_8, &
  217426565891488071_8,  164267268510338615_8,  393241344941049551_8,  248457607393726076_8, &
  718036578751400877_8,  309273214641235128_8,  346985850135213665_8, 1000764186787798137_8, &
  844130522682825875_8,  374151902974448954_8,  607758494955793995_8,  532672319999474291_8, &
  613969964055402732_8,  780410419194198488_8,  394895139131331966_8,  165455233759011797_8, &
    6322931772207497_8,   86161037070584406_8,  791443389714935878_8,  835759646653365969_8, &
  687059935766086720_8,   14214845501010438_8,  564352774506296125_8,  656139210352258808_8, &
  813986333089368255_8,  187942176056140984_8,   55182364551670509_8,  189085997088766303_8, &
  769961772185477576_8,  452568376836315185_8, 1032483914756410033_8,  733074542523761510_8, &
  725078382924804854_8,  370364211687393561_8,  211518325664242562_8,  198402155053317542_8, &
  701558338072560331_8,  685156488055193556_8,  972894316377842164_8,  618715301818574266_8, &
 1120461702831103893_8,  266035263849868872_8,  695159837979264076_8,  580660299994441057_8, &
   66777801238183396_8,  808242202788623316_8, 1131958431733874673_8,  143428505846952154_8, &
  478658011657139736_8,  424286141989330875_8,   97944871024137655_8,  127584227459937021_8, &
 1026980244582044584_8,   93529499555819729_8,  786072881728426537_8,  243198104306627423_8, &
   54461428243013698_8,  341821771035463176_8,  328610091866341975_8,  589596999450239151_8, &
   22268131370076300_8, 1125497851362970258_8,  338309767149066446_8,   48501308837176444_8, &
  506783924216248724_8,  550494901033541943_8,  853224896050219021_8,  540896285386052939_8, &
 1087963595273741451_8,  590236422416936651_8,  428199315726494527_8, 1017913446427102704_8, &
   11627268202307574_8,   12411272720587191_8,  151571015692077823_8,  801805729255844690_8, &
 1053129572400431340_8,  898149488598553006_8,   38812608295917889_8,  217051013754214696_8, &
 1027877295994496961_8,  243099290769625466_8,  613955609889484599_8,  353870262657925549_8, &
  637141720733912176_8, 1127408345606387682_8,  572192062258313139_8,  936847025720831096_8, &
  741596068164486248_8,  566216362278956981_8,  100840248057817976_8,  737232715384065277_8, &
 1082245789529037312_8,  788517729010872412_8, 1031957427233188981_8, 1104677670128566892_8, &
  432365578112025669_8, 1087453812475234390_8,  842879314247712024_8, 1091855776801052064_8/
data (mplog2con(i), i =   600,   799) / &
  278865056457498733_8,  274452313453978557_8,  908556800533554594_8,  577869998717774069_8, &
  670991146488849126_8, 1149310286936094747_8,  306699532648224706_8,  457514022108522018_8, &
  406707506328130099_8,  871051767037809673_8,  295783931878198179_8,  322158716553114590_8, &
  810383792479010363_8,  583197357626951498_8,  124697906935340833_8, 1051837863810602196_8, &
  772112765990823686_8,  211747207089338424_8,  799169051668571380_8, 1147020419962261891_8, &
  456292243478347382_8,  556793759222586816_8,  256638549248309571_8,  136321905145156104_8, &
  734776624627421328_8,  776166734497884968_8,  695577266205460858_8,  883556684626115256_8, &
  983723563201364370_8,  776786197748668391_8,  475552184055096133_8,  844266604857406075_8, &
  121615283512287317_8,    9067581013474256_8, 1021560708271259533_8,  906274065814214948_8, &
 1132245209448321733_8,  460342749845500015_8,  780415152079123787_8,  905996371999437994_8, &
  383819064556268321_8,  196051615838505789_8,  188741022513163602_8,  227986482207525482_8, &
  407229126723239000_8,  190085214224497956_8,  277296367059184976_8,  355820437457341298_8, &
 1078854378499317804_8,   32020025387782691_8,  640720685190603086_8,  852734663037663378_8, &
  264464739082922377_8,  335024594286679694_8, 1135663454385116605_8, 1120649339250429251_8, &
  394209779522642688_8,    2590995859022910_8,  682752152747476324_8,   94445426907114725_8, &
  664175841519509107_8,  200099227187547097_8,  980720931651783828_8,  175110786622650762_8, &
  326249008465240636_8,  222289956251610900_8,   84401765198866404_8,   43951738232093651_8, &
 1061998581558766947_8,  804016896777570374_8,  553987896409405118_8,  328381755879513320_8, &
  168077324838983330_8,  962664752752360065_8,  400567386176322048_8,  872842448194885666_8, &
  579472432376865602_8,  191824016414278687_8,   44792944137807009_8,  557072048817284581_8, &
  464780468776130630_8,  953713177036138577_8,  547498276706121232_8,  492659725118967205_8, &
  729853912709027694_8,  121906712729630341_8,  846479797716944612_8,  344666898443009521_8, &
  344359097202429800_8,  939005880601151541_8,  739469902012256961_8,  522699315080500473_8, &
  928611579704408774_8,   29824290460323223_8,  875823776529951743_8,  923035084868059011_8, &
  309221547396307487_8,  551955627778925301_8, 1144949620625652722_8,  208885703961551051_8, &
  857874732888983563_8,  142455790380519086_8,  796075857165951926_8,  186421529818299714_8, &
  986930304450790647_8,  875600270974391552_8, 1013234744964277088_8,  563065323890303076_8, &
  803630673386550924_8,  414190818950026776_8,  129105646808951793_8,  300998170897486172_8, &
  473454667902653769_8,  873512649259161579_8,   98472778980610450_8,  136551994856514226_8, &
 1077787456836457044_8,  840612906558582321_8, 1099970926062002697_8,  752497634121942619_8, &
  151512647715585905_8,  342868545578500791_8,  578889956118752508_8,  448735422218391621_8, &
  672062101866778817_8,  775940557852584381_8,     243985095924355_8,  271365311073457276_8, &
 1126841083685055294_8,  132604189951045671_8,  884730995621884774_8,  980685837546167370_8, &
  497367762513226619_8,  584582987052762645_8, 1104022329469484618_8,  697139211828998291_8, &
  415963849299775980_8,  240621556900763402_8,  832452481724583121_8, 1107969795684588140_8, &
  800041755863870066_8,  100851287044377902_8,  839058348490036754_8,  493920407171369266_8, &
  321652836031992975_8,  753445085031930560_8,  961245074392204540_8,  299400470291371394_8, &
  646378631354413115_8,  243818903187757550_8, 1022733408199555141_8,  139321989410550371_8, &
  548597010519710044_8,   83458605421088928_8,  501962253553281099_8,  464643749909362817_8, &
  195257349814776012_8, 1141292733263619832_8,  194278865892404084_8,  241466388992020488_8, &
  777347073769003549_8,  514876560340780717_8,  345470029516725675_8,  553372236299236517_8, &
  982422699855201039_8, 1005520980705024510_8,  510778287126193029_8,  225791943704232171_8, &
  731243211867607504_8,  278240161079565852_8,  967511878183205559_8,  988808584380813902_8, &
  789099488668750831_8,  915425111915078982_8,  488862429648177400_8,  823450492076598646_8, &
 1023499952444809049_8,   24927986448791912_8,  287862703120185408_8,  828608263965900141_8, &
  242140655783397369_8,  237589385998706936_8,  943808472644620496_8,  567924766317904788_8, &
  193435228639052855_8,   22499843949503154_8,  360703823095636586_8,   85863009663757278_8, &
  660109162120644625_8,  845928119477630086_8,  898001062188705721_8,  443980487767824843_8, &
  954148268409556975_8,  999864851324966066_8,  517177309607441607_8,  660794004363599433_8, &
  488234593502099210_8,  638320916434820525_8,  243359375222367005_8, 1020713578498329900_8/
data (mplog2con(i), i =   800,   999) / &
  171542867853781370_8,  107107949954094897_8,  774688289940185318_8,  232539131488437753_8, &
  488058626181754703_8,   30903178929158685_8,  188808334543967560_8,  211254432868961451_8, &
   78850033006999759_8,  798568116846010569_8,  848939684751205301_8,  566799850712825947_8, &
 1048167811024057127_8,  282948432002862356_8,  456461113592093132_8, 1040177238368624989_8, &
  517450413486254840_8,  806626212448774680_8,  998910289981461359_8,  470453719173931346_8, &
  269087033452990026_8,   79955433853542890_8, 1124633564761178700_8,  856853245839436641_8, &
  387174561729283646_8, 1052650223591992925_8,  280751504714819633_8,   29656364971427369_8, &
 1006082156511733776_8,  829128948265754398_8, 1059676261406298182_8,  228619449878404868_8, &
  809897153290506183_8, 1090339931415742412_8,  996607302723064978_8,  538870134414869278_8, &
  107585297277568183_8,  202666001776158248_8,  419409826462444037_8,  840811152998489919_8, &
  264903289087932283_8,  688078115437172311_8,  612726383633207051_8,  929374056782094999_8, &
   61607043280589062_8,  122169072554909443_8,  763125301038599055_8,  301081401927897956_8, &
  947058606381851692_8,  460238178716984995_8,  265878786938840815_8, 1069223716622168918_8, &
  113543438453979710_8,  217833552319080273_8,   23836046264100017_8,  879377680201552019_8, &
 1033850117437227616_8, 1008884643383989693_8,  582788836566860337_8,  870762109129573952_8, &
  150990451617713430_8, 1083582584353257121_8,  274048091708929711_8,  624890733022221610_8, &
  629386208685503491_8,   17161929256431812_8,  894223605435649381_8,  531668758695752063_8, &
  253911493609370662_8, 1134337565022246839_8,  421619976780838838_8, 1048120104303492877_8, &
  955832471347784992_8,  393451459371335945_8,  659836388821532813_8,  202550901643322383_8, &
  113411407396430508_8,  264815521189254049_8,  422447937075123938_8,  369708236503305351_8, &
  691364732958192854_8,  106560166810177108_8,  875549310068077107_8,  644063540984015438_8, &
  821864291168628043_8,  755402110936009514_8,  608148612990486740_8, 1032248994184851768_8, &
  983316556886450464_8,  724554986426719728_8,  747747654472932205_8,  167097846868702375_8, &
  315178218777542199_8, 1015169444386770642_8, 1058419236100997368_8,  949458617632420692_8, &
  211775190891291587_8,  410560258396857851_8,  497611518673367437_8,  693491450387497196_8, &
   61978781257529642_8,  842396480052797423_8,   21607201294239378_8,  688886848235344630_8, &
 1109432658327892577_8,  805986620017780112_8,  823543484053637107_8, 1134566853605974278_8, &
   83435206945411286_8,  751313197793774077_8,  406314023429654675_8,   25233395010318695_8, &
  756962578943244795_8,  537098983847573656_8,  596733154586351762_8,  462223637778679059_8, &
  633617075101491551_8,  432787930512540973_8,  423288589638437965_8, 1042658474804511529_8, &
  855417183730115965_8, 1074092783412956028_8,   97831489848130843_8,  406570034206291979_8, &
  737531051146394151_8,  865991417462274661_8,   90258971504800752_8,  202582896335699879_8, &
  414340105538112018_8,  467628253845031205_8, 1048114869176353655_8,  100261917520860596_8, &
  418409569780185631_8,  109308988914969830_8,  309097446051190109_8,  518273349308334253_8, &
  535038700955613132_8,  174851455450805754_8,   92975413975980655_8,  873811500227205444_8, &
  934510158897864617_8,  747636797957241269_8,  756714198212757108_8,  543822488904404678_8, &
  505608277523868598_8,  601922220038810633_8,  854151317755363235_8,  292530720454406209_8, &
  261200008100242161_8, 1100418728661175114_8,  238416850270270577_8,  847586693533435459_8, &
  492209755400529064_8,  913861962008259670_8,  788663778219619282_8,  604829301272092109_8, &
  345486168526580491_8,  842586687750480812_8,  453081855055969393_8,   88284284043064152_8, &
  724296532201234089_8,  973278548817724308_8,  576820060600330947_8,  623983914301568406_8, &
  220752079570261873_8,  706101464645326042_8,   39208276311132605_8, 1106578302091867316_8, &
  527381544245534032_8,  787270277233968762_8,  925728131504908658_8,  480794412548574332_8, &
  640807268486495588_8,  733165658843001879_8,  388666546765834172_8,  580155276576284938_8, &
  248629367540737029_8,  147696527560579820_8,  172236573393900306_8, 1116962164969714823_8, &
  656112047776515219_8,  860077860470590047_8,  267034879451351981_8,  734505304602729064_8, &
  334984534101030130_8,  379994277910803963_8,  118371934974875025_8,  395972233855319102_8, &
  483343081821137362_8,  695170314802749080_8,  475483276471859807_8,  457046897738984227_8, &
  646511902967665049_8,   23419384342660522_8, 1148251973969710269_8,  997084996784238994_8, &
  182129975172660425_8,   16107552625356663_8,  305769598179078383_8,  285372649627709787_8/
data (mplog2con(i), i =  1000,  1116) / &
  426154842883288599_8,  971820862552434137_8,  805034237090529991_8,  129381793112173426_8, &
  796349097178852575_8, 1095914385410367491_8,   60183289318960471_8,  584599920292003143_8, &
  891623033889996477_8,  335818849308490003_8,  338611762659913393_8, 1100499991732860217_8, &
  717570056436016399_8,  705222587167423071_8,  286517318764576461_8, 1031625786751779509_8, &
  725130176905804763_8,  958299491271527994_8,  393335852370618453_8,  712511833084678755_8, &
  995715193796251867_8, 1130301873578568392_8,  270196921674700807_8,  599909883133692107_8, &
  774578491820431175_8, 1108059265840577363_8,  366249482988655571_8, 1109218225939118509_8, &
  646926677756056721_8,  271098737163307748_8,  850662509776491758_8,  723107737018722679_8, &
  197335391718316535_8, 1058512337403515113_8,  667961857345747927_8,  932218810215668076_8, &
  689825524198451109_8,  347770987401131366_8,  811685514724324328_8,   24554820806102562_8, &
  978119655449061923_8,  750862400882654319_8,  898565190186309310_8,  488630659903647688_8, &
  338426341570688862_8,  548247867971520812_8,  305448011229652694_8,  604965641380790330_8, &
 1091814844674968957_8, 1135761340388512854_8,  291605092188706489_8,  724249218186247352_8, &
   75812089031399671_8,  571016455582880565_8,  606860802694100487_8,  372234225297475155_8, &
 1097366722261076015_8, 1031478345924192313_8, 1022316987518395207_8,    4466821150800125_8, &
  990319123187449648_8,  725423144304712913_8,  981258478983626423_8,  758262587395703761_8, &
 1025072375997053091_8,  789317478820524791_8,  883841627203700306_8,  545042391596686992_8, &
 1094203306398013013_8,  219085193191197938_8,  328054539967724316_8, 1121894228490578130_8, &
  542660567945635232_8,  854025998308892431_8,  348473564494718249_8,  499893383199186813_8, &
   77505978302706138_8,  291209354648696484_8, 1041548175314127155_8,  710376159574114493_8, &
  742459491848682217_8,  866912019902139358_8,  749873916030853074_8,  265506355397247338_8, &
  812238398374144870_8,   32877983561235972_8,  289572229816589031_8,  673341386514222320_8, &
  441583844367026154_8,  321316735962878817_8,    5446928668731077_8,  842593796470306992_8, &
  131917728081982580_8,  620962259354022480_8,  126931996390897249_8,  725451423698915638_8, &
  910983165055558525_8,  304653650992348508_8,  787842812933121661_8,  926219265248741290_8, &
  832322490588383093_8,  363036159374422479_8,  803568520393121849_8,  855832571969749538_8, &
  137003653649042314_8,   18111990899473286_8,  645309673747536544_8,   56841936167960499_8, &
  291710102143335432_8, 1028035077033195773_8,   50934365469588043_8, 1056350394205772163_8, &
  351398385142814178_8,  520906985190618220_8,  151310990670167219_8,                   0_8, &
                   0_8/
data (mppicon(i), i =     0,   199) / &
                1117_8,                1111_8,                1111_8,                   0_8, &
                   3_8,  163245215218020493_8,  221548887939444851_8,  309130526715976179_8, &
  130613387059588174_8,  488803641660957581_8,    5479683901017908_8, 1049557292833479548_8, &
  683510154637137333_8,  816401202565705053_8,  687214537561157899_8,  750286923632934871_8, &
  205987021816576225_8,  792306577480182695_8,  905300427237464356_8,  727776071756730224_8, &
  577009281997405206_8,  447705799768503526_8,  694778093217551165_8,  567692521958222059_8, &
  457240207437627925_8,  337461787920292901_8,  745826895745333546_8, 1091561318455640626_8, &
  603629619299580184_8,  832518932776066507_8,   27085472759848462_8,  629098671642603863_8, &
  558935848182577327_8,  215510146066288229_8,  370970457235231831_8,  326939946776609797_8, &
  417209043630362806_8,  814243075048349324_8, 1054216725227926249_8,  665194507147818747_8, &
  876561085637686296_8,  225018240481454520_8,  545813720412599148_8,  165777524090288146_8, &
  618548370491787416_8,  483217334367682177_8,  821388106713389065_8,  923152275941394378_8, &
  891815996563844996_8,  420486886110993858_8,  442200398449313827_8,  618092786094885968_8, &
 1111667774711939641_8,  207363804899066368_8,  318006183075913627_8,  423932541556190870_8, &
  908151578729491411_8,  616718980149489029_8,  301574693774108843_8,  365700002358539969_8, &
  250578582835556432_8,  571872118935852625_8,  952985216981846617_8,  281514769188966120_8, &
  439458784368426372_8,  746528506655796742_8, 1114034551826445120_8,  118300808647715428_8, &
 1068384875410585350_8,  126080960309080099_8,  952459468666309138_8,  328112955742002972_8, &
   43636676379639961_8,  123763628479188846_8,  639220755729488566_8,  546276656148253648_8, &
  342280970862940086_8,  291034183514835436_8,  151268767012813679_8,  789228094498091931_8, &
  210327034942942716_8,  365347592144604360_8,   91554887430375870_8, 1025976624634703830_8, &
  436611699830574003_8,  868273910521300800_8, 1139986159296435195_8,  987792021002495491_8, &
  147165924834033708_8,  515546132858793905_8, 1126969244813752539_8,  225795174353169919_8, &
  964169533966261960_8,  779216377417554783_8,  791828173830697339_8,  325349948703499717_8, &
  557261559359871623_8,  388134139240234324_8,  191673637953061804_8,  464764015041926966_8, &
  674176209240348872_8, 1017807150032814362_8,    4778742094045571_8,  829981519211846941_8, &
  241956690100598520_8,  312539654364316863_8,  835295555666238116_8,  916450871061066780_8, &
 1073201314967636698_8,  245293073184582127_8, 1021174859005549530_8,  349848146367392472_8, &
 1040731271587745934_8,  944923576184473827_8,  890357640240347023_8, 1111878655948273224_8, &
  614943739001827356_8,  358834321438080049_8,  923748035026200769_8,  779952807247339745_8, &
  538156469497793282_8,  142459710325455702_8, 1113770200545875662_8,  620979545570606607_8, &
  942343825727437697_8,  948810973079534118_8,  468538699673750643_8,   90654326887705298_8, &
   28560596123699028_8,  301564951776788156_8,  986020566887145686_8,  293223511351616656_8, &
   10430161118933854_8,  154952024968629770_8, 1091107195437713849_8,  136610166817005050_8, &
  480329081725507930_8,  376035209373163566_8,  412578007950259810_8,  674689545354188420_8, &
 1041257584893832650_8,  554306231371894789_8,  187553770414411607_8,  287044812409058826_8, &
  321587250192189626_8,  503089730398846761_8,  498891487096411574_8,  447635503846497135_8, &
 1095506307798357604_8,  376898111334381818_8,   74514210531214599_8,  478709388540599090_8, &
  667899940469392409_8,  171764430523794586_8,  567310191039658127_8, 1070491593718941462_8, &
  691891878188307052_8,  876823062741475370_8,  393230765185325331_8,  292243059471807817_8, &
  619908240383503247_8, 1030596023304415773_8,  188034011416819021_8,  203663015593499668_8, &
  926932860427889446_8,  448159412085976261_8, 1011454401153252079_8,  906289439368984231_8, &
   45432366818415264_8, 1020183486342912677_8,    2031064390573183_8, 1002866518089027269_8, &
  510862482797681207_8, 1082272113827627504_8,  288794099193875701_8,  118161591024900151_8, &
  745587670108721553_8,   89957413876099395_8,  164226505403466042_8, 1035361805052784076_8, &
  627537691398102439_8,  762311290272940080_8, 1075455169109718302_8,  296713713305763488_8, &
 1022091168921491518_8,  807011733229241590_8,  988830867127075830_8,   45119502442113282_8, &
  330869958218723442_8,  849764824281360241_8, 1142105643364354990_8,   84949316114141479_8, &
  152891951348469786_8,  999352682484889509_8,  596472362293663676_8,  718794519288993998_8, &
  899290813149027834_8,  446019324575370301_8,  211418598108960217_8,   39921290235235371_8/
data (mppicon(i), i =   200,   399) / &
  726788873862195813_8,  884136031320809779_8,  784995716227964432_8,  222054448217161061_8, &
 1085677138754230001_8,  971155150171345283_8,  817725098599550950_8, 1070026160813501355_8, &
 1093281886101210851_8,  312734842145675360_8, 1035276347048113015_8,  143677328572051362_8, &
  677190008698248463_8,  577560939938081922_8,  373320247313961139_8,  480866260400819758_8, &
  754535440346979850_8,  206971173125911393_8,  948919325033893657_8,  875833601340108943_8, &
  509771759670394478_8, 1057274947006350425_8,  407148917456731959_8,  941279603527004973_8, &
 1004230614513444184_8,  150965741334572780_8, 1055267121758812970_8, 1078571247209740353_8, &
  369776928024387952_8, 1066742295767413073_8,    5573047304806845_8,  675596362291257398_8, &
  274637609371716720_8,  730600574065622202_8,  783086401075199072_8,  442138337876495035_8, &
  212659217084653231_8,  115931460912952987_8,  677047079540910170_8,  226238447255493102_8, &
  708683767450409812_8,   49988703351314783_8,  547607790105766136_8,  250101140378802033_8, &
  138929200125448833_8,   63992137707839997_8,  474811862802258106_8,  690598135051956775_8, &
  172152201001507522_8,  317904937376724274_8, 1022410092215592045_8,  847541519710629619_8, &
  344243806291946849_8, 1074042296520188237_8,  713931281892773981_8,   90639208443829998_8, &
  195523587261258604_8,  356171040914476148_8,  212868678448559461_8,  356824320067113053_8, &
  585038723373419975_8, 1028145036595303973_8,   58373767874448678_8,  319694585981574267_8, &
  478365714219231691_8,  524990930598540649_8,  194364009170796717_8,  113867422509504322_8, &
  252779923729714474_8,   55555430459039555_8, 1018727692615512816_8,  660016179483338238_8, &
  932645492131331294_8,  249277698126775304_8,  428589419758513830_8,   33786075147471342_8, &
  643971132962740021_8,  762967038553538668_8,  169453072701983690_8,  993576189139104851_8, &
  325664166075764254_8,  715119851185451269_8,  398794410895768035_8,  465053889218168407_8, &
   89571543984175262_8, 1087302590553907742_8,  544847857231258527_8,  197516521671398357_8, &
  750012620818871884_8,  171639565760807953_8,  369567255670093756_8, 1100763891159271053_8, &
  290522244573099114_8,  304975162816595482_8, 1089371889376633925_8,  315606106708375736_8, &
  594560457388168885_8,  656108340838741610_8,   13306013701237763_8,  853845923628244227_8, &
  129343269068548019_8,  387260726082950270_8,  480699932129906808_8,  168093623794049736_8, &
  493964078889003228_8,   91810820904715904_8, 1063030002180550223_8,  288206386754148555_8, &
  399940785366423222_8,  769238022069908990_8,  922807617703668522_8,  297818022576234296_8, &
  412622773079980818_8,  353728131878581430_8,  943598845884248554_8, 1024985783258294093_8, &
  962436829150308343_8,  911840453717962580_8, 1069884958790991958_8,  777896270352438579_8, &
  736937238415801963_8,  764342041144767834_8,  605660074942814412_8,  761836461189083962_8, &
  444249247537243641_8,  427639524689089303_8,  936177972912858927_8,  506667814354965122_8, &
  928750604686779590_8,  491495493339446384_8, 1118823484261789249_8,   18560804323826245_8, &
  975150365234123243_8,  962006759963872280_8, 1099178705066228600_8,  235878958679064462_8, &
  622375685383196178_8,  814061014909315830_8,  454812756416206187_8,  160262629597195795_8, &
  802779153118968098_8,  602121296111444185_8,  711131189388092151_8,  183804429027859321_8, &
  359554033433896908_8, 1107609825559640183_8,  957290638892410867_8,  281705739472745497_8, &
  661747069931712171_8, 1103258572185218502_8,   45576809841499105_8,  346133848303364021_8, &
  364911859707881317_8,  955816234341912338_8,  652397200042599349_8,  256979122796810422_8, &
  467142890219077020_8,  723141035467822570_8,  268829697630100397_8,  164303628632810429_8, &
  498552902490770449_8,  402957827878518594_8, 1022648516208338886_8,  478142083905146055_8, &
  589673705573073463_8,  826732819264445374_8,  453732591445109520_8,  108399403435545179_8, &
  997796776742831170_8,  100645588407184784_8,  931241086950189668_8,  789947348731725307_8, &
 1062824809100264446_8,  710381292173097992_8,  459499635835143008_8,  350471359457878913_8, &
 1130534647625996086_8, 1138504932131028737_8, 1057063540428474172_8,    1660225950415851_8, &
 1002242388162658969_8,  861738828415995720_8, 1149224936363291887_8,  254585989755123311_8, &
  703626640314643573_8, 1091371323874842542_8,  803436833661748356_8,  477636205185949220_8, &
  463406278634657648_8,  634256963748769252_8,  916776836296049157_8,  937059360910563146_8, &
  693193149181780137_8,  990957269172231235_8,  163875509483012617_8, 1084216066669412945_8, &
  967990315052514589_8,   52453636623068960_8, 1092238295709824986_8,  591505990259187244_8/
data (mppicon(i), i =   400,   599) / &
 1052522681944133649_8,  423367435060488704_8,  195277722846504858_8, 1119358325704123420_8, &
  243277970169923253_8, 1081276925563852910_8,     105933928717437_8,  441667678763103118_8, &
  748534059479909058_8,  100006890760923403_8,  916164679351247310_8,  401338678932029588_8, &
  827538116120283709_8,  561023343976658725_8, 1113037397724676443_8,  814370791516722501_8, &
  394074848599857213_8,  562163642904887781_8,   67376635378596002_8,  599345535363658694_8, &
 1141037877017267908_8,  243946432642182275_8,  565089818679337858_8,   82877817785943552_8, &
   63114414730499066_8, 1107035294448973228_8,  372984429052141135_8,  735058901755655177_8, &
  701606885883743901_8,  860459320219980231_8, 1064820680785628187_8,  482476804831651563_8, &
  125654621631043400_8,  770955328130358378_8,  249618161836673107_8,   71933230724278237_8, &
  392668713470414278_8,  185663248834208869_8,   48577690340318947_8,   20863977760662229_8, &
  115514620963363462_8, 1072468076759057458_8,  301449414113995420_8, 1093541589881854889_8, &
  858334085349971280_8,  839435060255025711_8,  695588692014377365_8,  607692750239593214_8, &
 1140527066372521732_8,  501754982910236686_8,  335823319742153115_8,   44590503809669438_8, &
  689329308630833047_8,  202393142171973813_8,  114513776228793766_8,  566640962104182738_8, &
  948042489956118002_8,  831395362660361640_8,  645506647253086688_8,  115508526105415646_8, &
  978401091602597005_8,  181187957687670370_8,  682597907491741441_8,  657199360099973472_8, &
 1028197111911159045_8,  494631314286944314_8,   99401371877424579_8, 1061126937193577001_8, &
  515410286110967594_8,  122713029722124831_8,  429421305597975795_8,  114651839437935967_8, &
 1005713760570215098_8,  273508487079730208_8,  781929427076474828_8,  781697743960765715_8, &
  590214054562060386_8,  662609245495601420_8,  214834108693553614_8,  452458912770853742_8, &
  251853211068418270_8,  375652825896061482_8, 1011076659558884457_8,  599787787263284843_8, &
  259596938973135069_8,  398232034620869770_8, 1034630511187417229_8,  291833613596932942_8, &
  269052292399774297_8, 1149636040661945291_8,  346167435919796474_8,  452695019793492715_8, &
 1097874763438351459_8,  379135634950596465_8,  861962567683411469_8,  639077601100400417_8, &
  464345009716281408_8,  914376870519948109_8,  163934989336937697_8,  792706692096958928_8, &
  484312558848614472_8,  501011955760368312_8,  144425376169816615_8, 1118600618699553075_8, &
 1142994265260635205_8,  169969141059294492_8, 1042947454800607136_8,  142897607827294187_8, &
  920935506785669571_8,  729172662926510196_8,  708278717168802526_8,  917978605756101532_8, &
  228213058120945145_8,  222873253531942777_8, 1034059091971619570_8,  959724059444402635_8, &
 1123404204793592874_8,   70122209191477586_8,  430909135531194012_8, 1061770368333720490_8, &
  635489962604552965_8,  477593822701182128_8,  262408212101074454_8,  688403907114306531_8, &
  178148803656783154_8, 1012528360912978754_8,  831320890927680116_8,  294064605113375719_8, &
   81279705719299893_8,  718563531623165536_8,  823955390406401509_8,  346114448956426524_8, &
 1036524797657251439_8,   99513595260486096_8,  397600802542648098_8,  693363299071976231_8, &
  442359539068655308_8,  923284992293217703_8,  426634074140350344_8,  947793101666527528_8, &
  124211623711772699_8,  512395926215355515_8,  950956697818882512_8,   30468737336122021_8, &
  287494425167313701_8,  864167055961581329_8,  173676322760041142_8,  917880883236160181_8, &
  221026851445859283_8,  255222856736340703_8,   52416058391742158_8,  533530295502243670_8, &
   27168708955638504_8,  638815889316840655_8,  696507739442142210_8, 1132840849344212557_8, &
  498740645747554409_8,  749231317999194706_8,  991364654590021217_8,  228812210073598205_8, &
 1144823341402994746_8,  914268108348696300_8,  503096937643564364_8,  861855622251829283_8, &
 1077741448626252528_8, 1040767569161345410_8,    8807764107699497_8,  840095131873034120_8, &
  456769090322611276_8, 1072243966380053006_8,  322151313524637155_8,  773987976946509093_8, &
  987056743911026298_8,  493643932638416845_8,  660125598750882950_8, 1119860095173264420_8, &
 1075175910445779404_8,  513354668034299888_8,  442122892089012611_8,  347361177477359662_8, &
  264247774697419514_8,  201174223151248813_8,  140186384052173025_8,  357450420738028801_8, &
  335066712757483082_8,  305764990370978606_8,  292867629567733648_8,  711791125801957234_8, &
  869863439542552344_8,  886783735113342972_8, 1045841156130879185_8,  196835175534531835_8, &
   51532982163929509_8,  259390229499896424_8,  335551754210047672_8,  298903780632381541_8, &
  368124299874320777_8,  959624407079722323_8, 1029145421298505455_8,  318071778594678750_8/
data (mppicon(i), i =   600,   799) / &
 1071297798809568407_8,  518992790295221002_8,  616936618661776078_8, 1060760059633173168_8, &
 1017521720810691726_8, 1043573831805115145_8,  489181487301733752_8,  641732161361706498_8, &
  790033570088026856_8,  971367919752476560_8, 1050511046983989958_8, 1151367868361306003_8, &
  151919057777764520_8, 1067952769870011321_8,  710065525415532004_8,  341936976146628828_8, &
  362825178883977810_8,  275596201093612068_8, 1060138530090755105_8,   43966104131745934_8, &
  981642510550314903_8,  539118837775256522_8,  232758249934238083_8,  591161896580232746_8, &
  975875067442073077_8,  934516619388227943_8, 1082381477451456600_8,  622951457146967118_8, &
 1106503024727354785_8,  567400702122704059_8, 1046932860717838812_8,  938257555016460148_8, &
 1097285986120910461_8,  262249223347657593_8,  657132871808198372_8,  372830029715726824_8, &
  325168415081831683_8,  104475676958710564_8,  569149818378726201_8, 1087772481220225532_8, &
  992416400572751862_8,  690444477756277008_8,  100613230703668065_8,  953931084086685434_8, &
  625564345307896989_8,  813867927041913424_8,  902192666140827540_8,  647720439546040181_8, &
  136448434079390154_8,  723734538083509480_8,  337840266306612852_8,  722825919187517635_8, &
  504362442033402712_8,  837219949044583649_8,  914893802568537045_8,  801587043494028111_8, &
  394388132612581804_8,  386930410710233184_8,  296399048304558789_8,  260133673861225909_8, &
  850066080844741999_8,  122248522314685408_8, 1035994194036708682_8,  467060437931421398_8, &
  438423778608438674_8,  660017893582929523_8,  892747664873971525_8,  551713310068024908_8, &
  950428783514874250_8,  632543407514153832_8, 1111397300078008665_8,  174139869594430607_8, &
  935482183007308732_8,  140867884308043868_8,  565899223729731533_8,  468283228320275276_8, &
  581009777062092197_8,  124727587478005720_8,     957693647855020_8,  727931981101339393_8, &
  744353006463908431_8,  827803025595318430_8,  393099827202063966_8,  573365264589170861_8, &
  370467711105595481_8,  706400161932801028_8,  523306761183789525_8,  140496182440362109_8, &
  602654892638241360_8,  854273130926188695_8, 1071894502734959163_8,  727605128467842919_8, &
  586597537542562517_8,  641634222948277344_8,   40745008904822933_8,  115812574431699933_8, &
  932226805940754753_8,  674240375350886297_8,  247714204353199290_8, 1071456624787147855_8, &
  645210083095010568_8,   66320281592083992_8,  600648595949770448_8,  955671375291449992_8, &
  410423156555247693_8,  319602796669768711_8,  812675259004306178_8,  475909500491494344_8, &
  905517779594044918_8,  898940397308968190_8,  495034836871257826_8,  727123010021176901_8, &
  767113063022743327_8,  904591403831536975_8,   91786549998231265_8,  132216142609666622_8, &
 1094249398817683068_8,  308515899861853897_8, 1142418996837768150_8, 1109644164580479236_8, &
  851699040903447341_8,  938271765022509176_8,   10598230849177549_8,  775201908538417177_8, &
  895057427772715819_8,   48278546484144369_8, 1049267729932011133_8, 1123400876065663929_8, &
  860256900888456639_8,  992871634825421964_8,  929272593477481867_8,  896427123266261437_8, &
   36427989990237761_8,  416237530499580845_8,  461761675978468651_8,  840399478064667520_8, &
  395769322899414131_8,  802398258443108679_8,  518563240307624497_8, 1088518482954630019_8, &
  738151565092561879_8,  150034689867971597_8,  931199397573581479_8,  538704065636387145_8, &
   71187567004261412_8,  445264343163873388_8,  652351489650769333_8, 1092773657793363046_8, &
 1003735675027378821_8,   94048941308685914_8,  435943727121749832_8,  596873906601242850_8, &
  408388480839662965_8,  911355713430166159_8,  606431818365418355_8,  874904933863273650_8, &
  323676684129919256_8,   64150339000351190_8,  182592830486849383_8, 1032652832708465984_8, &
  582782088904130992_8,  967586551128196963_8,  484538236543982368_8,  463865476535547039_8, &
 1082122655787315500_8,   64824798333089183_8,   83242840952983173_8,  830986217466038223_8, &
  483193039926455592_8,  665944119556224225_8,   84850718232333231_8,  605303142101472638_8, &
  187267547820190628_8,    7508796889304933_8,  400569341598891134_8,  691748288114339749_8, &
  504989974784528163_8,   38508400151925637_8,  470369294719372532_8,  116232796450372843_8, &
  849536437153813566_8,   77983885405741242_8,   36631071043607096_8,  898707438408351751_8, &
 1063037910526942402_8,  611661644155963484_8,   15352133684695822_8,  769056555843724319_8, &
  792693057289603771_8,  397360794136385185_8,  779242029728093169_8,  310139790379785256_8, &
  180860703526638733_8,   21749322919107328_8,  519302453209073774_8, 1138895943637119413_8, &
   38711698915228100_8,  719993131856928946_8,  790578719940331802_8, 1083100657525698823_8/
data (mppicon(i), i =   800,   999) / &
  435921859080409804_8,  683859700350302683_8,  224408844934079738_8,   64062858055166832_8, &
  557906537221420365_8,  515881141598943277_8,  413750453503730423_8,   51322907665264401_8, &
  964696595234198504_8,  232592790637985813_8, 1102877689642802589_8,  137464178751582378_8, &
  485837969167564451_8,    7061694381170294_8,  656613522942125872_8,   71405825786382508_8, &
  228111626328775022_8,  370342730481644567_8, 1118464641136427772_8,  809396842274320332_8, &
  968747591202856048_8,  545572763726792532_8,  114022411656886261_8,  750715224150815279_8, &
   75361024281878212_8, 1103183575411308858_8,  317494798189727881_8,  726211830069553748_8, &
  848392752301587238_8,  515837948444824953_8,  500067048703696521_8, 1137986868193745443_8, &
  600978865535978095_8,  561631987733100716_8,  304346817339418554_8,  551744473249252034_8, &
  696672977504067279_8,  862252903688135965_8,  181077465951216997_8,   13816519969146727_8, &
  596870961726389703_8,  370337548752297798_8,  101271545835159262_8,   76090136150498517_8, &
 1013086142418967041_8,  300273600868137964_8,  655016978566379688_8,  195839269992104576_8, &
  674983923591804060_8,  804994011941200673_8,  325890055801692942_8,  165846561741549189_8, &
  942921890419373924_8, 1002919343762028755_8,  146813462418537765_8,  754533952790608511_8, &
  681188917026544196_8,  861851659122373904_8,  337173256059144399_8,   20338778714613003_8, &
 1035579202838301492_8, 1104526176869871042_8,  560900020457962407_8,  859766629653011101_8, &
  546149027385075786_8,   55941649597302435_8,  284019497721383241_8,   53128470725581510_8, &
  306213346829030544_8,  737850965785264078_8,  501376169270697070_8, 1070932720730191216_8, &
  156427127977381302_8,  878545584687233630_8,  714423476050270602_8,  857520761046882670_8, &
  151876277936961589_8, 1152057274154960867_8,  813360104868515020_8,  246076687664095957_8, &
  756271332343014605_8,  666741679751712501_8,  916496972024055849_8,  887164017613808388_8, &
 1022544735925390345_8,  866030006550386156_8,  433437555791625681_8,  392115846733673325_8, &
  838167602632271011_8,  265963724909522856_8,  518367335995547107_8,  862341900825383830_8, &
    3822883952496752_8,  899504826901541836_8,   99124922061942241_8,   99286404248652475_8, &
  963951179419665801_8,  722321218269613703_8,  582396237363530859_8,  713207056714163153_8, &
   87572713904423990_8,  751530305399325468_8,  469830716649184948_8,  258635271568531384_8, &
  717193941637982838_8,  818257497761904645_8,  913446176381770860_8, 1110387277686847617_8, &
  355073235246730091_8,  987168387766052359_8,  199365090686538939_8,  461606783862269064_8, &
  395640949531528740_8,  611886440921089174_8,  612199002052626802_8, 1100888007597249181_8, &
  184967050614576171_8, 1081108654420748584_8,  934217096057883641_8,   72527755808771929_8, &
  197793939771527617_8,  836825914087774701_8,  588568863675067147_8,  977227188649590203_8, &
  652842651161834092_8,  108439542847846691_8, 1049916010542913886_8, 1072932622883929081_8, &
  545457576712216839_8,  910148148247261345_8,  412171042383460579_8,  643963586267682386_8, &
  693349082513547396_8, 1025723163039371207_8, 1099869253417316512_8,  193825348936926842_8, &
  991672070874240695_8,  832104773549101279_8,  299036973101102924_8,  478730593115693954_8, &
  528855334168795017_8,  207826200297386623_8,  169233128223864120_8,  384995516814316025_8, &
  240480589962934588_8, 1041274463433539864_8,  327055291618690086_8, 1052260099830717042_8, &
  961846811081948106_8, 1098401844393556361_8,  615353155778303215_8,  512633924258508306_8, &
  819061000620201966_8,   97278518409877913_8,  574259243111728771_8,  914708573921563556_8, &
 1122196832129288617_8,  600624207280723695_8,  234708564322242556_8,  331885530727851390_8, &
  568138828831247153_8,  515803018093970379_8,  469977572478525816_8,  638255583681584054_8, &
  270209934941068610_8,  731280974035176986_8,  878106903318701483_8,  535468971829388992_8, &
  701959402678572657_8, 1086308637726031787_8,  914294756369509759_8,  475183880396342175_8, &
 1086510848495192633_8,  353753591940524302_8,  917037902115425609_8, 1043414896378705581_8, &
  855226531306980668_8,  632579447240364063_8,  907700399667344235_8,  971185566741072906_8, &
  874750547894182625_8,  575483041571081879_8,  909402777220934393_8, 1083501709858291387_8, &
 1096585993983468443_8,  139704801708340731_8,  284001923813218699_8,  171473015951168066_8, &
  185596178009345321_8,  861902696686557701_8, 1080837709014526528_8, 1077634080909673371_8, &
  653437640175444861_8,  228807947704407030_8,  592098790577148796_8,  890067626935423168_8, &
  441537495795765413_8, 1060671578553127197_8, 1074207413264856389_8,  434639518912097367_8/
data (mppicon(i), i =  1000,  1116) / &
   91502086598018387_8, 1055686388652374320_8,  557190800486253063_8,  815439364490083386_8, &
   84966373018511719_8,  276995072069481162_8,  220526751575517413_8,  972260112228629558_8, &
 1085521452294961529_8,  713027274959290285_8, 1151516579614649520_8,  660366430030084098_8, &
  460660278415423450_8,  454290926768519759_8,  944170765619993852_8, 1093819913056384742_8, &
  242311117027896833_8,  569777999001083134_8,  281300663479999602_8,  656176045708108361_8, &
  864399413967615071_8,  781863448895587594_8,  222424586086972759_8,  851412490298422381_8, &
 1127281275731160733_8,  406541390120931870_8,  752259664563816305_8,   72705718391831624_8, &
  236686916035121382_8, 1111809480857157650_8,  602199136978509852_8,  684847631953943878_8, &
 1141132800083243126_8, 1126979914804451964_8,  560176389796043205_8,  113167092493977553_8, &
  907862631694763588_8, 1078044950378582581_8, 1106013951213258845_8,  846418050553168074_8, &
  192386549681314049_8,  844950000941166953_8,  247728967959939618_8,  458636325193080767_8, &
  671542013742435065_8,  102993857964978993_8,   39150035991658758_8, 1152660213212165215_8, &
  290334734983060691_8,  735904953723145947_8,  104032781304093246_8,  249831666172095710_8, &
    3154090832205159_8, 1109732315068353250_8,  144644884193264485_8,   70214458438594779_8, &
  832326974180709953_8, 1070981139929216902_8,  792475385880950006_8,  155375900487983189_8, &
  154167715879877523_8, 1110642083395673325_8,  599777367207642692_8,  563246372342768852_8, &
 1021900770567761848_8, 1141508483825033964_8,  915100124885560396_8,  961242267149702421_8, &
  750248093756832303_8,  895257227468808830_8, 1074771164705836571_8,  121619851849489802_8, &
 1142214762738312275_8,  256492664737530673_8,  617375037921663163_8,  509350389270748757_8, &
 1135804246176955933_8,  777497502310497120_8,  630172658511656645_8,  378915530885345103_8, &
  377730987277820196_8,  512732864286521778_8,  384560264588369116_8,  674904731320476002_8, &
  870295562703256141_8,  603332192725907287_8,  922245186974062365_8,  594450130570465346_8, &
  632093663824517706_8, 1093188729195098730_8, 1010470117276801206_8, 1062685030377708486_8, &
  732010324935348387_8, 1147732280164422973_8,  616174040908811045_8,  421595017181802079_8, &
   65752240039950205_8,  131322276630060995_8, 1100059463454792783_8,  174570284100925599_8, &
   50680903182113732_8,  848996591539105372_8,  830800796531764614_8,  368387844562805881_8, &
  557769924498866829_8,  589651025514406575_8,  817121523055458575_8,  767084900178280219_8, &
  616038697526650331_8,  458126119918118820_8,  536288130006954897_8,  952969870740457625_8, &
  466921250634065894_8,  637809219400435158_8,   39839887081268338_8,                   0_8, &
                   0_8/

end module mpfuna


